#!/usr/bin/env python
# Rui Zhang 7.2020
# rui.zhang@cern.ch

import os
from argparse import ArgumentParser
import importlib
import numpy as np
from datetime import datetime
import glob

import pandas as pd
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import rcParams
rcParams.update({'figure.autolayout': True})
import seaborn as sns

import quple

# Expressibility (arXiv: 1905.10876)
def calculate(args):
    nqubits = args.nqubits
    PQCnames = args.PQCs
    output = args.output_folder
    repeats = args.repeats
    plotting = args.plotting
    nbins = args.nbins
    statistics = args.statistics

    rows_list = []
    for nqubit in nqubits:
        for PQCname in PQCnames:
            folder = f'{output}/qu{nqubit}/pqc{PQCname}'
            if not os.path.exists(folder):
                os.makedirs(folder)
            for repeat in repeats:
                print('Evaluate', PQCname, repeat)
                if PQCname == 'FirstOrderExpansion':
                    encoder = importlib.import_module(f'quple.data_encoding.first_order_expansion')
                    cq = encoder.FirstOrderExpansion(feature_dimension=nqubit, parameter_symbol='data', depth=repeat)
                elif PQCname == 'SecondOrderExpansion':
                    encoder = importlib.import_module('quple.data_encoding.second_order_expansion')
                    cq = encoder.SecondOrderExpansion(feature_dimension=nqubit, parameter_symbol='data', depth=repeat)
                elif PQCname == 'QConvNN':
                    QNN = importlib.import_module(PQCname)
                    qnn = QNN.QuantumNeuralNetwork(nqubits = nqubit, use_quple = True, nlayers = repeat)
                    qnn.get_quantum_model(nqubit)
                    cq = qnn.m_circuit
                else:
                    QNN = importlib.import_module(PQCname)
                    qnn = QNN.QuantumNeuralNetwork(nqubits = nqubit, use_quple = True, nlayers = repeat)
                    qnn.get_quantum_model(nqubit)
                    cq = qnn.m_circuit
                print(cq)
                for i in range(statistics):
                    dic = {
                        'Nqubits': nqubit,
                        'PQC': PQCname,
                        'Nrepeat': repeat,
                        'absExpressibility':  0,
                        'relExpressibility':  0,
                        'MeyerWallach': 0,
                        'vonNeumann': 0,
                        'GradientVariance': 0,
                    }
                    dic['absExpressibility'] = quple.circuit_expressibility_measure(cq, samples=1000, bins=nbins)
                    dic['relExpressibility'] = -np.log(dic['absExpressibility'] / ((2**nqubit-1)*(np.log(nbins))))
                    dic['MeyerWallach'] = quple.circuit_entangling_measure(cq, samples=500)
                    dic['vonNeumann'] = quple.circuit_von_neumann_entropy(cq, samples=500)
                    if plotting:
                        plt = quple.circuit_fidelity_plot(cq, samples=3000, bins=nbins)
                        name = f'{folder}/fidelity-repeat{repeat}-bin{nbins}-i{i}.pdf'
                        plt.savefig(name)
                        plt.clf()
                        print(f'\n\033[92m{datetime.now().strftime("%Y-%m-%d-%H-%M-%S")} Save fidelity plot: {name}\033[0m')
                    rows_list.append(dic)
                    df = pd.DataFrame(rows_list)
                    rp = '_'.join([str(i) for i in repeats])
                    name = f'{folder}/inspect-repeat{rp}-bin{nbins}-stat{statistics}.csv'
                    df.to_csv(name, index=False)
                    print(f'\n\033[92m{datetime.now().strftime("%Y-%m-%d-%H-%M-%S")} Save inspect dataframe: {name}\033[0m')
    return df


def main(args):
    if not args.load:
        calculate(args)

    else:
        output = args.load
        nqubits = args.nqubits

        xTitle, yTitles, lTitle = 'PQC', ['absExpressibility', 'relExpressibility', 'MeyerWallach', 'vonNeumann'], 'Nrepeat'
        xName, yNames, lName = 'Parameterised Quantum Circuit', ['Expressibility', 'Relative Expressibility', 'Meyer-Wallach Entanglement', 'von Neumann Entanglement'], 'N depths'

        for nqubit in nqubits:
            # extract all jobs
            sub_folders = glob.glob(f'{output}/qu{nqubit}/pqc*/*.csv')

            frames = []
            for sub_folder in sub_folders:
                print(f'\033[92mRead from: {sub_folder}\033[0m')
                pqc = sub_folder.split('/')[-2]
                names = sub_folder.split('/')[-1].split('-')
                nbin = names[-2][3:]
                statistics = names[-1].split('.')[0][4:]
                frames.append(pd.read_csv(sub_folder))
            df = pd.concat(frames)

            sub_df = df.loc[df['Nqubits']==nqubit]
            sub_df = sub_df.sort_values('PQC')

            color = 'RdBu'
            for yTitle, yName in zip(yTitles, yNames):
                fig, ax = plt.subplots(tight_layout=True)
                chart = sns.boxplot(x=xTitle, y=yTitle, hue=lTitle, data=sub_df, palette=color, fliersize=0)
                chart.set_xticklabels(chart.get_xticklabels(), rotation=10)
                # jitter plot overlay
                sns.stripplot(x=xTitle, y=yTitle, hue=lTitle, data=sub_df, jitter=True, palette=color, alpha=0.8, dodge=True, linewidth=1, edgecolor='gray').set(xlabel=xName, ylabel=yName)
                handles, labels = ax.get_legend_handles_labels()
                plt.legend(handles[0:int(len(handles)/2)], labels[0:int(len(handles)/2)], title = lName, framealpha=0.5)

                filename = f'{output}/qu{nqubit}/{yTitle}.pdf'
                plt.savefig(filename)
                print(f'\n\033[92mSave file: {filename}\033[0m')


if __name__ == '__main__':

    """Get arguments from command line."""
    parser = ArgumentParser(description="Evaluate Parametrised Quantum Circuits.")

    parser.add_argument('--load', action='store', default=None, help='Load dataframe file (default: %(default)s)')
    parser.add_argument('--output-folder', action='store', default='../output/inspect', help='Output folder for logs (default: %(default)s)')
    parser.add_argument('--PQCs', nargs='+', type=str, choices=['FirstOrderExpansion', 'SecondOrderExpansion', 'QConvNN', 'QDensNN', 'QZinnerCNOTNN', 'QEfficientSU2NNQNN', 'QMNISTNN'], help='PQC type')
    parser.add_argument('--repeats', nargs='+', type=int, help='Repeating time (default: %(default)s)')
    parser.add_argument('--nbins', type=int, default=500, help='Number of qubits / varibles (default: %(default)s)')
    parser.add_argument('--nqubits', nargs='+', type=int, help='Number of qubits / varibles (default: %(default)s)')
    parser.add_argument('--plotting', action='store_true', default=False, help='Plotting fidelity (default: %(default)s)')
    parser.add_argument('--statistics', type=int, default=10, help='Number evaluation times (default: %(default)s)')

    main(parser.parse_args())



#!/usr/bin/env python
# Rui Zhang 7.2020
# rui.zhang@cern.ch

import logging

import tensorflow as tf
import tensorflow_quantum as tfq
import numpy as np
import sympy
import cirq

def add_output_layer(circuit, output, gate, prefix):
    for i, qubit in enumerate(circuit.all_qubits()):
        if output == qubit:
            continue
        symbol = sympy.Symbol(prefix + '-' + str(i))
        circuit.append(gate(qubit, output)**symbol)

class CircuitLayerBuilder():
    def __init__(self, data_qubits, readout):
        self.data_qubits = data_qubits
        self.readout = readout
    
    def add_layer(self, circuit, gate, prefix):
        for i, qubit in enumerate(self.data_qubits):
            symbol = sympy.Symbol(prefix + '-' + str(i))
            circuit.append(gate(qubit, self.readout)**symbol)
'''
Develop your own Quantum Neural Network and make a link to the class at the bottom:
QuantumNeuralNetwork = YourClassName
'''
class QEfficientSU2NNQNN(object):
    def describe(self): return self.__class__.__name__
    def __init__(self, **kargs):
        
        self.m_nqubits = kargs['nqubits']
        self.m_loss = kargs['loss']
        self.m_use_quple = kargs['use_quple']
        self.m_nlayers = kargs['nlayers']

        logging.getLogger(__name__)

    def create_quantum_model(self, size):
        """Create a QNN model circuit and readout operation to go along with it."""
        data_qubits = cirq.GridQubit.rect(size[0], size[1]) 
        circuit = cirq.Circuit()

        entangler_map= [[i, i + 1] for i in range(size[1] - 1)]

        param_idx = 0
        num_parameters = (size[1]+1)*size[1]*2

        symb = sympy.symbols('θ_0:'+str(num_parameters))
        symbols = np.array(symb)
        symbols = symbols.reshape(size[1]+1, size[1], 2)

        for i, qubit in enumerate(data_qubits):
            circuit.append(cirq.ry(symbols[0][i][0])(qubit))
            circuit.append(cirq.rz(symbols[0][i][1])(qubit))
            param_idx += 2

        for d in range(self.m_nlayers):
            for src, targ in entangler_map:
                circuit.append(cirq.CNOT(data_qubits[src],data_qubits[targ]))
            for i, qubit in enumerate(data_qubits):
                circuit.append(cirq.ry(symbols[d+1][i][0])(qubit))
                circuit.append(cirq.rz(symbols[d+1][i][1])(qubit))
                param_idx += 2

        readout = cirq.GridQubit(-1, -1) # a single qubit at [-1,-1]
        circuit.append(cirq.H(readout))
        add_output_layer(circuit, readout, cirq.YY, 'yy')
        add_output_layer(circuit, readout, cirq.ZZ, 'zz')
        return circuit, cirq.Z(readout)

    def get_quantum_model(self, input_size):
        # Build the model circuit
        model_circuit, model_readout = self.create_quantum_model(size=(1, input_size))
        self.m_circuit = cirq.Circuit()
        self.m_circuit.append([model_circuit, model_readout])
        # SVGCircuit(model_circuit)
        logging.info(f'model circuit:\n{model_circuit}')

        # Build the Keras model
        model = tf.keras.Sequential(
            [
                # The input is the data-circuit, encoded as a tf.string
                tf.keras.layers.Input(shape=(), dtype=tf.string),
                # The PQC layer returns the expected value of the readout gate, range [-1,1].
                tfq.layers.PQC(model_circuit, model_readout),
            ]
        )
    
        if self.m_loss == 'hinge':
            def hinge_accuracy(y_true, y_pred):
                y_true = tf.squeeze(y_true) > 0.0
                y_pred = tf.squeeze(y_pred) > 0.0
                result = tf.cast(y_true == y_pred, tf.float32)
                return tf.reduce_mean(result)
                
            model.compile(loss=tf.keras.losses.Hinge(),
                optimizer=tf.keras.optimizers.Adam(learning_rate=0.01, ),
                metrics=[hinge_accuracy],
                )
            
        model.summary(print_fn=logging.getLogger(__name__).info)

        return model

QuantumNeuralNetwork = QEfficientSU2NNQNN


#!/usr/bin/env python
# Rui Zhang 7.2020
# rui.zhang@cern.ch

import logging

import tensorflow as tf
import tensorflow_quantum as tfq
import sympy
import cirq

from quple.trial_wavefunction.real_amplitudes import RealAmplitudes

def add_output_layer(circuit, output, gate, prefix):
    for i, qubit in enumerate(circuit.qubits):
        symbol = sympy.Symbol(prefix + '-' + str(i))
        circuit.append(gate(qubit, output)**symbol)

class CircuitLayerBuilder():
    def __init__(self, data_qubits, readout):
        self.data_qubits = data_qubits
        self.readout = readout
    
    def add_layer(self, circuit, gate, prefix):
        for i, qubit in enumerate(self.data_qubits):
            symbol = sympy.Symbol(prefix + '-' + str(i))
            circuit.append(gate(qubit, self.readout)**symbol)
'''
Develop your own Quantum Neural Network and make a link to the class at the bottom:
QuantumNeuralNetwork = YourClassName
'''
class QDensNN(object):
    def describe(self): return self.__class__.__name__
    def __init__(self, **kargs):
        
        self.m_nqubits = kargs['nqubits']
        self.m_use_quple = kargs['use_quple']
        if self.m_use_quple:
            self.m_nlayers = kargs['nlayers']

        logging.getLogger(__name__)

    # Now build a two-layered model, matching the data-circuit size, and include the preparation and readout operations.
    def create_quantum_model(self, size):
        circuit, readout = None, None
        if self.m_use_quple:
            '''Create a QNN model circuit and readout operation to go along with it.'''
            circuit = RealAmplitudes(n_qubit=self.m_nqubits, copies=self.m_nlayers, parameter_symbol='model', final_rotation_layer=False, entangle_strategy='cyclic')

            readout = cirq.GridQubit(-1, -1) # a single qubit at [-1,-1]
            circuit.append(cirq.H(readout))

            add_output_layer(circuit, readout, cirq.ZZ, 'zz')
        else:
            data_qubits = cirq.GridQubit.rect(size[0], size[1])  # match data-circuit defined in `convert_to_circuit`
            readout = cirq.GridQubit(-1, -1) # a single qubit at [-1,-1]
            circuit = cirq.Circuit()
            builder = CircuitLayerBuilder(
                data_qubits = data_qubits,
                readout=readout)
            # Then add layers (experiment by adding more).
            builder.add_layer(circuit, cirq.ZZ, "zz1")

        # Finally, prepare the readout qubit.
        circuit.append(cirq.H(readout))
        return circuit, cirq.Z(readout)

    def get_quantum_model(self, input_size):
        # Build the model circuit
        model_circuit, model_readout = self.create_quantum_model(size=(1, input_size))
        self.m_circuit = cirq.Circuit()
        self.m_circuit.append([model_circuit, model_readout])
        # SVGCircuit(model_circuit)
        logging.info(f'model circuit:\n{model_circuit}')

        # Build the Keras model
        model = tf.keras.Sequential(
            [
                # The input is the data-circuit, encoded as a tf.string
                tf.keras.layers.Input(shape=(), dtype=tf.string),
                # The PQC layer returns the expected value of the readout gate, range [-1,1].
                tfq.layers.PQC(model_circuit, model_readout),
            ]
        )

        def hinge_accuracy(y_true, y_pred):
            y_true = tf.squeeze(y_true) > 0.0
            y_pred = tf.squeeze(y_pred) > 0.0
            result = tf.cast(y_true == y_pred, tf.float32)
            return tf.reduce_mean(result)
            
        model.compile(loss=tf.keras.losses.Hinge(),
            optimizer=tf.keras.optimizers.Adam(),
            metrics=[hinge_accuracy],
            )
        model.summary(print_fn=logging.getLogger(__name__).info)

        return model

QuantumNeuralNetwork = QDensNN


#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

from argparse import ArgumentParser
import glob
import numpy as np
import json

from sklearn.metrics import auc
from sklearn.metrics import plot_roc_curve

import matplotlib.pyplot as plt
from matplotlib import rcParams
import matplotlib
rcParams.update({'figure.autolayout': True})
font = {'size': 22}
matplotlib.rc('font', **font)


def averageROC(top_folder, group_folders, catagories, plot_individual=True, skip=True):
    print('catagories', catagories)
    plt.figure()
    for jobfeature in sorted(group_folders):
        folders = group_folders[jobfeature]
        tprs_x = np.linspace(0, 1, 100)
        fprs_y, aucs = [], []
        Traintype, NNtype = catagories[0]
        plt.grid(color='gray', linestyle='--', linewidth=1, alpha=0.3)
        for i, folder in enumerate(folders):
            logfile = glob.glob(f'{folder}/result/roc*.json')[-1]


            with open(logfile, 'r') as json_file:
                print('read', logfile)
                roc_curves_to_save = json.load(json_file)

            fpr, tpr, roc_auc = roc_curves_to_save[NNtype][Traintype]
            fpr_y, tpr_x, roc_auc = 1.0 - np.array(fpr), np.array(tpr), np.array(roc_auc)
            interp_fpr_y = np.interp(tprs_x, tpr_x, fpr_y)
            interp_fpr_y[0] = 1.0
            fprs_y.append(interp_fpr_y)
            aucs.append(roc_auc)

            if plot_individual:
                plt.plot(tpr_x, fpr_y, alpha=0.5, lw=1, label=f'{Traintype} {NNtype} AUC ({i+1}): {roc_auc: 0.3f}')
            
        mean_fprs_y = np.mean(fprs_y, axis=0)
        mean_fprs_y[-1] = 0.
        mean_auc = auc(tprs_x, mean_fprs_y)
        std_auc = np.std(aucs)

        plt.plot(tprs_x, mean_fprs_y, color='b', label=r'Mean ROC (AUC = %0.2f $\pm$ %0.2f)' % (mean_auc, std_auc), lw=2, alpha=.8)

        std_fprs_y = np.std(fprs_y, axis=0)
        fprs_y_upper = np.minimum(mean_fprs_y + std_fprs_y, 1)
        fprs_y_lower = np.maximum(mean_fprs_y - std_fprs_y, 0)
        plt.fill_between(tprs_x, fprs_y_lower, fprs_y_upper, color='grey', alpha=.2, label=r'$\pm$ 1 std. dev.')

        plt.plot([0, 1], [1, 0], linestyle='--', color='black', label='Luck')
        plt.xlim([-0.05,1.05])
        plt.ylim([-0.05,1.05])
        plt.xlabel('Signal acceptance')
        plt.ylabel('Background rejection')
        plt.legend(loc='lower left')
        roc_file = f'{top_folder}/roc_average_{jobfeature}.pdf'
        print('Save', roc_file)
        plt.savefig(roc_file, format='pdf')
        plt.clf()

from extracAUC import extractJobFeatures
def main(args):
    input_folder = args.input_folder
    catagory = args.catagory
    catagories = [(i, j) for j in ['QNN'] for i in catagory]

    for j, top_folder in enumerate(input_folder):
        # extract all jobs
        sub_folders = glob.glob(f'{top_folder}/2020*')

        # extract job features
        group_folders = extractJobFeatures(sub_folders)
        averageROC(top_folder, group_folders, catagories)


if __name__ == '__main__':

    """Get arguments from command line."""
    parser = ArgumentParser(description="Plot AUC information.")

    parser.add_argument('-i', '--input-folder', action='store', nargs='+', default=[], help='Folder of the log files (default: %(default)s)')
    parser.add_argument('-c', '--catagory', action='store', nargs='+', default=['Test'], help='Events to plot (default: %(default)s)')
    
    main(parser.parse_args())



#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

import logging
import os
import numpy as np

from datetime import datetime
import collections 
import json, pickle
from importlib import reload
import importlib

from sklearn.utils import shuffle
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import roc_curve, auc
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')

class AutoVivification(dict):
    '''Implementation of perl's autovivification feature.'''
    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value

class _train(object):
    def describe(self): return self.__class__.__name__
    def __init__(self, additiona_log, **kargs):
        self.m_input_file = kargs['input_file']
        self.m_output_folder = kargs['output_folder']
        self.m_log_file = kargs['log_file']
        self.m_training_size = kargs['training_size']
        self.m_test_size = kargs['test_size']
        self.m_val_size = kargs['val_size']
        self.m_nqubits = kargs['nqubits']
        self.m_epochs = kargs['epochs']
        self.m_batch_size = kargs['batch_size']
        self.m_fix_seed = kargs['fix_seed']
        self.m_test_predict_QNN = None
        self.m_test_predict_DNN = None
        self.m_train_predict_QNN = None
        self.m_train_predict_DNN = None
        self.m_val_predict_QNN = None
        self.m_val_predict_DNN = None
        self.padding = kargs['padding']
        self.m_plot_input = kargs['plot_input']
        self.m_plot_input_only = kargs['plot_input_only']
        self.m_reducer = 'PCA'
        self.m_loss = kargs['loss']
        self.m_large_pca = kargs['large_pca']
        self.m_entangle = kargs['entangle']
        self.m_entangle_gate = kargs['entangle_gate']
        self.m_additiona_log = additiona_log
        
        self.m_log_dir = {}
        self.m_log_dir['base'] = os.path.join(self.m_output_folder, 'logs_' + self.m_log_file, f'{datetime.now().strftime("%Y-%m-%d-%H-%M-%S")}-{self.m_fix_seed}-{os.uname()[1]}-bs{self.m_batch_size}-qu{self.m_nqubits}-lpca{self.m_large_pca}-train{self.m_training_size}-red{self.m_reducer}-{self.m_additiona_log}')
        if self.m_plot_input or self.m_plot_input_only:
            self.m_log_dir['input'] = os.path.join(self.m_log_dir['base'], 'input')
        if not self.m_plot_input_only:
            self.m_log_dir['result'] = os.path.join(self.m_log_dir['base'], 'result')
        self.m_log_dir['model'] = os.path.join(self.m_log_dir['base'], 'model')

        for folder in self.m_log_dir.values():
            if not os.path.exists(folder):
                os.makedirs(folder)

        logging.info(f'Logger folder is: {self.m_log_dir["base"]}')
        self.config_logging()
        logging.info(f'Train option {vars(self)}')
        
    def config_logging(self):
        log_filename = f'train{self.m_training_size}_val{self.m_val_size}_test{self.m_test_size}_bs{self.m_batch_size}.log'
        self.m_log_filename = os.path.join(self.m_log_dir["base"], log_filename)

        reload(logging)
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s",
            handlers=[
                logging.FileHandler("{0}".format(self.m_log_filename)),
                logging.StreamHandler()
            ])

    def plot_input(self, training_signal, training_backgd, test_signal = None, test_backgd = None, column = 0, name = 'var', norm = False):
        _, nbins, _ = plt.hist(training_signal[:, column], histtype='step', bins = 100, linewidth=2, alpha=0.9, label='Train Signal', density=norm)
        plt.hist(training_backgd[:, column], histtype='step', bins = nbins, linewidth=2, alpha=0.9, label='Train Backgd', density=norm)
        if test_signal is not None and test_signal.shape[0]: plt.hist(test_signal[:, column], histtype='step', bins = nbins, linewidth=2, alpha=0.9, label='Test Signal', linestyle = 'dashed')
        if test_backgd is not None and test_backgd.shape[0]: plt.hist(test_backgd[:, column], histtype='step', bins = nbins, linewidth=2, alpha=0.9, label='Test Backgd', linestyle = 'dashed')

        plt.ylabel('Events')
        plt.xlabel(name)
        plt.legend(loc='best')
        plt.tight_layout(pad=0)
        plt.savefig(f'{self.m_log_dir["input"]}/{name}.pdf')
        logging.debug(f'Save column {column} to {self.m_log_dir["input"]}/{name}.pdf')
        plt.clf()

    def load_npz_data(self):
        '''
        load npz data

        :Returns:
            training_input, test_input, each of which is a dictionary.
            keys = 'A', 'B'
            values = nEvent x nVariable
        ''' 
        data = np.load(self.m_input_file, allow_pickle=True)
        if isinstance(data, np.lib.npyio.NpzFile):
            training_input = {'input':v for k,v in np.ndenumerate(data['training_input'])}['input']
            test_input = {'input':v for k,v in np.ndenumerate(data['test_input'])}['input']
            if self.m_training_size:
                training_input = {key: training_input[key][:self.m_training_size] for key in training_input.keys()}
            if self.m_test_size:
                test_input = {key: test_input[key][:self.m_test_size] for key in test_input.keys()}
            val_input = None
        elif isinstance(data, np.ndarray):
            if self.padding != -999:
                data[data == -999] = self.padding
                logging.info(f'Change padding to {self.padding} from default -999.')

            signal_array = data[data[:, -1] > 0.5][:, :-1]
            backgd_array = data[data[:, -1] < 0.5][:, :-1]
            np.random.shuffle(signal_array)
            np.random.shuffle(backgd_array)

            # If not specified, 1/2 training, 1/4 for val and 1/4 for test
            if not self.m_training_size and not self.m_test_size and not self.m_val_size:
                self.m_training_size = int(signal_array.shape[0] / 2)
                self.m_test_size = int(signal_array.shape[0] / 4)
                self.m_val_size = self.m_test_size
            # If training specified, 1/2 for val and 1/2 for test
            elif self.m_training_size and not self.m_test_size and not self.m_val_size:
                self.m_test_size = int((signal_array.shape[0] - self.m_training_size) / 2)
                self.m_val_size = self.m_test_size
            # If training & test specified, val = min(rest, test)
            elif self.m_training_size and self.m_test_size and not self.m_val_size:
                self.m_val_size = min([m_test_size, signal_array.shape[0] - self.m_training_size - self.m_val_size])
            # If training & val specified, test = rest
            elif self.m_training_size and not self.m_test_size and self.m_val_size:
                self.m_test_size = signal_array.shape[0] - self.m_training_size - self.m_val_size
            elif self.m_training_size and self.m_val_size and self.m_test_size:
                pass
            else:
                logging.fatal(f'Request of training size {self.m_training_size}, test size {self.m_test_size} and val size {self.m_val_size} from {len(signal_array)} signal or {len(backgd_array)} bkg events do NOT make sense.')
                assert(False)
            if self.m_training_size + self.m_test_size + self.m_val_size > signal_array.shape[0]:
                logging.fatal(f'Request of training size {self.m_training_size}, test size {self.m_test_size} and val size {self.m_val_size} from {len(signal_array)} signal or {len(backgd_array)} bkg events do NOT make sense.')
                assert(False)

            logging.info(f'Request of training size {self.m_training_size}, test size {self.m_test_size} and val size {self.m_val_size} from {len(signal_array)} signal or {len(backgd_array)} bkg events make sense.')
            training_input, test_input, val_input = {}, {}, {}
            pca_input = {}
            training_input['1'], training_input['0'] = signal_array[:self.m_training_size], backgd_array[:self.m_training_size]
            test_input['1'], test_input['0'] = signal_array[self.m_training_size:self.m_training_size+self.m_test_size], backgd_array[self.m_training_size:self.m_training_size+self.m_test_size]
            if self.m_val_size == 0:
                val_input = None
            else:
                val_input['1'], val_input['0'] = signal_array[self.m_training_size+self.m_test_size:self.m_training_size+self.m_test_size+self.m_val_size], backgd_array[self.m_training_size+self.m_test_size:self.m_training_size+self.m_test_size+self.m_val_size]
            pca_input['1'], pca_input['0'] = signal_array, backgd_array
        
        name = 'BeforeDimRed_{}'
        if self.m_plot_input:
            for column in range(training_input['0'].shape[1]):
                self.plot_input(training_input['1'], training_input['0'], test_input['1'], test_input['0'], column = column, name = name.format(str(column)))
            logging.info(f'Save {training_input["0"].shape[1]} input columns to {name.format("*")}.pdf')
        elif self.m_plot_input_only:
            for column in range(training_input['0'].shape[1]):
                self.plot_input(signal_array, backgd_array, column = column, name = name.format(str(column)), norm = True)
            logging.info(f'Save {training_input["0"].shape[1]} input columns to {name.format("*")}.pdf')
        else:
            logging.info(f'Save not {training_input["0"].shape[1]} input columns to {name.format("*")}.pdf')

        logging.info(f'Signal: {training_input["1"][:3]}')
        logging.info(f'Backgd: {training_input["0"][:3]}')

        return training_input, test_input, val_input, pca_input

    def process_npz_data(self, dictAB, training=False):
        '''
        Process dictionary from npz file:
            - merge and shuffle training data in two classes
            - MinMax scale X
        '''
        target = {'1': 1, '0': -1} if self.m_loss == 'hinge' else {'1': 1, '0': 0} # 1=signal; 0=backgd
        X = np.concatenate((dictAB['0'], dictAB['1']), axis=0)
        y = np.array([target['0']]*dictAB['0'].shape[0] + [target['1']]*dictAB['0'].shape[0])

        dimension = self.m_nqubits + self.m_entangle
        if self.m_nqubits < X[0].shape[0]:
            ''' Do Dimension Reduction '''
            reducer_file = f'{self.m_log_dir["model"]}/reducer_{self.m_reducer}.pkl'
            if training:
                ''' Perform Dimension Reduction '''
                module = importlib.import_module(f'sklearn.decomposition')
                reducer = module.PCA(n_components=dimension).fit(X)
                X = reducer.transform(X)
                pickle.dump(reducer, open(reducer_file, 'wb'))
                logging.info(f'Write reducer to: {reducer_file}, with dimension {dimension}, X shape is {X.shape}')
            else:
                ''' Apply Dimension Reduction '''
                reducer = pickle.load(open(reducer_file,'rb'))
                X = reducer.transform(X)
            ''' Plot input '''
            name = 'AfterDimRed_' + self.m_reducer + '_{}_' + ('training' if training else 'test')
            if self.m_plot_input or self.m_plot_input_only:
                for column in range(X.shape[1]):
                    self.plot_input(X[y == target['1']], X[y == target['0']], column = column, name = name.format(str(column)))
                logging.info(f'Save {X.shape[1]} input columns to {name.format("*")}.pdf')
            else:
                logging.info(f'Save not {X.shape[1]} input columns to {name.format("*")}.pdf')

        elif self.m_nqubits == X[0].shape[0]:
            ''' No Dimension Reduction '''
            logging.info(f'No Dimension Reduction needed as dimension {self.m_nqubits} is same as input.')
        else:
            logging.fatal(f'Request of Dimension Reduction dimension {self.m_nqubits} is greater than input dimension {X[0].shape[0]}.')

        scaler_file = f'{self.m_log_dir["model"]}/scaler.pkl'
        scale_range = (-1, 1)
        X, y = shuffle(X, y)
        if training:
            scaler = MinMaxScaler(feature_range=(scale_range))
            scaler.fit_transform(X, y)
            X = scaler.transform(X)
            pickle.dump(scaler, open(scaler_file, 'wb'))
            logging.info(f'Write scalar to: {scaler_file} with range {scale_range}')
        else:
            scaler = pickle.load(open(scaler_file, 'rb'))
            logging.info(f'Load scalar from: {scaler_file}')
            X = scaler.transform(X)

        name = 'AfterTransf_{}' + ('_training' if training else '_test')
        if self.m_plot_input or self.m_plot_input_only:
            for column in range(X.shape[1]):
                self.plot_input(X[y == target['1']], X[y == target['0']], column = column, name = name.format(str(column)))
            logging.info(f'Save {X.shape[1]} input columns to {name.format("*")}.pdf')
        if self.m_plot_input_only:
            exit(0)
        else:
            logging.info(f'Save not {X.shape[1]} input columns to {name.format("*")}.pdf')

        return X, y

    def get_classical_data(self):
        training_input, test_input, val_input, pca_input = self.load_npz_data()
        if self.m_large_pca:
            _, _ = self.process_npz_data(pca_input, training = True)
            x_train, y_train = self.process_npz_data(training_input)
        else:
            x_train, y_train = self.process_npz_data(training_input, training = True)
        x_test, y_test = self.process_npz_data(test_input)
        x_val, y_val = self.process_npz_data(val_input) if val_input else (x_test, y_test)

        logging.info(f'Training size: {len(x_train)}, target: {y_train[:10]}...')
        logging.info(f'Testing size: {len(x_test)}, target: {y_test[:10]}...')

        self.m_data = collections.namedtuple('data', ['x_train', 'y_train', 'x_test', 'y_test', 'x_val', 'y_val'])(x_train, y_train, x_test, y_test, x_val, y_val)


    def plotLoss(self, history, title):
        plt.plot(history.history['loss'])
        plt.plot(history.history['val_loss'])
        plt.title(title)
        plt.xlabel('Epochs')
        plt.ylabel('Loss')
        plt.legend(['train', 'val'], loc='upper right')
        plt.savefig(f'{self.m_log_dir["result"]}/loss_{title}.pdf', format='pdf')
        plt.clf()

    def getAUC(self, y_test, weight, score):
        fpr, tpr, _ = roc_curve(y_test, score, sample_weight=weight)
        roc_auc = auc(fpr, tpr)
        return fpr, tpr, roc_auc

    def plotROC(self, excludeDNN=False):
        plt.figure()
        plt.grid(color='gray', linestyle='--', linewidth=1)
        
        curves = []
        Curve = collections.namedtuple('Curve', ['y', 'predict', 'NNtype', 'Traintype', 'Linestyle'])
        if self.m_test_predict_QNN is not None:  curves.append(Curve(y=self.m_data.y_test, predict=self.m_test_predict_QNN, NNtype='QNN', Traintype='Test', Linestyle='solid'))
        if self.m_train_predict_QNN is not None: curves.append(Curve(y=self.m_data.y_train, predict=self.m_train_predict_QNN, NNtype='QNN', Traintype='Train', Linestyle='dashed'))
        if self.m_val_predict_QNN is not None: curves.append(Curve(y=self.m_data.y_val, predict=self.m_val_predict_QNN, NNtype='QNN', Traintype='Val', Linestyle='dashed'))
        
        if not excludeDNN:
            if self.m_test_predict_DNN is not None:  curves.append(Curve(y=self.m_data.y_test, predict=self.m_test_predict_DNN, NNtype='DNN', Traintype='Test', Linestyle='solid'))
            if self.m_train_predict_DNN is not None: curves.append(Curve(y=self.m_data.y_train, predict=self.m_train_predict_DNN, NNtype='DNN', Traintype='Train', Linestyle='dashed'))
            if self.m_val_predict_DNN is not None: curves.append(Curve(y=self.m_data.y_val, predict=self.m_val_predict_DNN, NNtype='DNN', Traintype='Val', Linestyle='dashed'))

        logging.info(f'ROC curves: {len(curves)}.')

        roc_curves_to_save = AutoVivification()
        for curve in curves:
            fpr_test, tpr_test, roc_auc_test = self.getAUC(curve.y, None, curve.predict)
            roc_curves_to_save[curve.NNtype][curve.Traintype] = (fpr_test.tolist(), tpr_test.tolist(), roc_auc_test.tolist())
            fpr_test = 1.0 - fpr_test
            plt.plot(tpr_test, fpr_test, label=f'{curve.Traintype} {curve.NNtype} AUC: {roc_auc_test: 0.3f}', linestyle=curve.Linestyle)
            logging.info(f'{curve.Traintype} {curve.NNtype} AUC values: {roc_auc_test} from {len(curves)} curves.')
        
        with open(f'{self.m_log_dir["result"]}/roc{len(curves)}_train{len(self.m_data.y_train)}_val{len(self.m_data.y_val)}_test{len(self.m_data.y_test)}.json', 'w') as f:
            json.dump(roc_curves_to_save, f)

        plt.plot([0, 1], [1, 0], linestyle='--', color='black', label='Luck')
        plt.xlim([-0.,1.])
        plt.ylim([-0.,1.])
        plt.xlabel('Signal acceptance')
        plt.ylabel('Background rejection')
        plt.legend(loc='lower left')
        roc_file = f'{self.m_log_dir["result"]}/roc{len(curves)}_train{len(self.m_data.y_train)}_val{len(self.m_data.y_val)}_test{len(self.m_data.y_test)}.pdf'
        plt.savefig(roc_file, format='pdf')
        plt.clf()



#!/usr/bin/env python
# Rui Zhang 8.2020
# rui.zhang@cern.ch

import logging
import numpy as np
from math import ceil
import tensorflow as tf
import tensorflow_quantum as tfq
import sympy
import cirq

'''
Develop your own Quantum Neural Network and make a link to the class at the bottom:
QuantumNeuralNetwork = YourClassName
'''
class QConvNNHYB(object):
    def describe(self): return self.__class__.__name__
    def __init__(self, **kargs):
        
        self.m_nqubits = kargs['nqubits']
        self.m_single_gates = [cirq.X, cirq.Y, cirq.Z]
        self.m_double_gates = [cirq.ZZ, cirq.YY, cirq.XX]
        self.m_use_quple = kargs['use_quple']
        if self.m_use_quple:
            self.m_nlayers = kargs['nlayers']

        logging.getLogger(__name__)

    def cluster_state_circuit(self, bits):
        """Return a cluster state on the qubits in `bits`."""
        circuit = cirq.Circuit()
        circuit.append(cirq.H.on_each(bits))
        for this_bit, next_bit in zip(bits, bits[1:] + [bits[0]]):
            circuit.append(cirq.CZ(this_bit, next_bit))
        return circuit

    def one_qubit_unitary(self, bit, symbol):
        """Make a Cirq circuit enacting a rotation of the bloch sphere about the X,
        Y and Z axis, that depends on the values in `symbols`.
        """
        return cirq.Circuit([gate(bit)**sympy.Symbol(symbol + '_gate' + str(i)) for i, gate in enumerate(self.m_single_gates)])

    def two_qubit_unitary(self, bits, symbol):
        """Make a Cirq circuit that creates an arbitrary two qubit unitary."""
        circuit = cirq.Circuit()
        circuit += [self.one_qubit_unitary(bits[i], symbol+'_before'+str(i)) for i in range(2)]
        circuit += [gate(*bits)**sympy.Symbol(symbol+'_gate'+str(i)) for i, gate in enumerate(self.m_double_gates)]
        circuit += [self.one_qubit_unitary(bits[i], symbol+'_after'+str(i)) for i in range(2)]
        return circuit

    def two_qubit_pool(self, source_qubit, sink_qubit, symbol):
        """Make a Cirq circuit to do a parameterized 'pooling' operation, which
        attempts to reduce entanglement down from two qubits to just one."""
        pool_circuit = cirq.Circuit()
        sink_basis_selector = self.one_qubit_unitary(sink_qubit, symbol+'_sink')
        source_basis_selector = self.one_qubit_unitary(source_qubit, symbol+'_sour')
        pool_circuit.append(sink_basis_selector)
        pool_circuit.append(source_basis_selector)
        pool_circuit.append(cirq.CNOT(control=source_qubit, target=sink_qubit))
        pool_circuit.append(sink_basis_selector**-1)
        return pool_circuit

    def quantum_conv_circuit(self, bits, symbol):
        """Quantum Convolution Layer following the above diagram.
        Return a Cirq circuit with the cascade of `two_qubit_unitary` applied
        to all pairs of qubits in `bits` as in the diagram above.
        """
        circuit = cirq.Circuit()
        for adjacence in zip(bits[0::2], bits[1::2] + [bits[0]]):
            circuit += self.two_qubit_unitary(adjacence, symbol)
        for adjacence in zip(bits[1::2], bits[2::2] + [bits[0]]):
            circuit += self.two_qubit_unitary(adjacence, symbol)
        return circuit

    def quantum_pool_circuit(self, bits, symbol):
        """A layer that specifies a quantum pooling operation.
        A Quantum pool tries to learn to pool the relevant information from two
        qubits onto 1.
        """
        circuit = cirq.Circuit()
        length = len(bits)
        for source, sink in zip(bits[:int(length/2)], bits[ceil(length/2):]):
            circuit += self.two_qubit_pool(source, sink, symbol)
        return circuit, bits[ceil(length/2):]

    def create_quantum_model(self, qubits):
        """Create sequence of alternating convolution and pooling operators 
        which gradually shrink over time."""
        model_circuit = cirq.Circuit()
        # Cirq uses sympy.Symbols to map learnable variables. TensorFlow Quantum
        # scans incoming circuits and replaces these with TensorFlow variables.
        for i in range(self.m_nlayers):
            model_circuit += self.quantum_conv_circuit(qubits, 'clay'+str(i))
            if i==1: break
            circuit, qubits = self.quantum_pool_circuit(qubits, 'play'+str(i))
            model_circuit += circuit
        # Create our qubits and readout operators in Cirq.
        readout = qubits[-1]
        return model_circuit, cirq.Z(readout)


    def get_quantum_model(self, size):

        # Build a sequential model enacting the logic in 1.3 of this notebook.
        # Here you are making the static cluster state prep as a part of the AddCircuit and the
        # "quantum datapoints" are coming in the form of excitation
        cluster_state_bits = cirq.GridQubit.rect(1, size)
        excitation_input = tf.keras.Input(shape=(), dtype=tf.dtypes.string)
        cluster_state = tfq.layers.AddCircuit()(excitation_input, prepend=self.cluster_state_circuit(cluster_state_bits))
        
        # Build the model circuit
        model_circuit, model_readout = self.create_quantum_model(cluster_state_bits)
        self.m_circuit = cirq.Circuit()
        self.m_circuit.append([model_circuit, model_readout])
        logging.info(f'model circuit:\n{model_circuit}')

        quantum_model = tfq.layers.PQC(model_circuit, model_readout)(cluster_state)
        # qcnn_model = tf.keras.Model(inputs=[excitation_input], outputs=[quantum_model])

        dense = tf.keras.layers.Dense(1)(quantum_model)
        qcnn_model = tf.keras.Model(inputs=[excitation_input], outputs=[dense])

        # Custom accuracy metric.
        @tf.function
        def custom_accuracy(y_true, y_pred):
            y_true = tf.squeeze(y_true)
            y_pred = tf.map_fn(lambda x: 1.0 if x >= 0 else -1.0, y_pred)
            return tf.keras.backend.mean(tf.keras.backend.equal(y_true, y_pred))

        qcnn_model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=0.01, ), loss=tf.losses.mse, metrics=[custom_accuracy]) #TODO zhangr lr
        qcnn_model.summary(print_fn=logging.getLogger(__name__).info)
        return qcnn_model

    def generate_data(self, qubits):
        """Generate training and testing data."""
        n_rounds = 20  # Produces n_rounds * n_qubits datapoints.
        excitations = []
        labels = []
        for n in range(n_rounds):
            for bit in qubits:
                rng = np.random.uniform(-np.pi, np.pi)
                excitations.append(cirq.Circuit(cirq.rx(rng)(bit)))
                labels.append(1 if (-np.pi / 2) <= rng <= (np.pi / 2) else -1)

        split_ind = int(len(excitations) * 0.7)
        train_excitations = excitations[:split_ind]
        test_excitations = excitations[split_ind:]

        train_labels = labels[:split_ind]
        test_labels = labels[split_ind:]

        return tfq.convert_to_tensor(train_excitations), np.array(train_labels), tfq.convert_to_tensor(test_excitations), np.array(test_labels)

QuantumNeuralNetwork = QConvNNHYB



#!/usr/bin/env python
# Rui Zhang 7.2020
# rui.zhang@cern.ch

import tensorflow as tf

def get_quantum_model(num_var):
    # A simple model based off LeNet from https://keras.io/examples/mnist_cnn/
    cmodel = tf.keras.Sequential(
        [
            tf.keras.layers.Dense(num_var, activation='relu'),
            tf.keras.layers.Dense(10, activation='relu'),
            # tf.keras.layers.Dropout(0.5),
            tf.keras.layers.Dense(1, activation='sigmoid'),
        ]
    )

    cmodel.compile(loss=tf.keras.losses.BinaryCrossentropy(from_logits=True),
        optimizer=tf.keras.optimizers.Adam(),
        metrics=['accuracy'],
        )
    
    return cmodel



#!/usr/bin/env python
# Rui Zhang 7.2020
# rui.zhang@cern.ch

import os
import re
from argparse import ArgumentParser
import glob
from math import ceil

import pandas as pd
from pprint import pprint
import matplotlib.pyplot as plt
from matplotlib import rcParams
rcParams.update({'figure.autolayout': True})
plt.rcParams["figure.figsize"] = (15,10)
import seaborn as sns
import datetime

process_map = {
    'hmumu': r'$H  \rightarrow \mu \mu$',
    'ttH': r'$t\bar{t}H$',
}
Encoder_map = {
    'FirstOrderExpansion': '1st-Order',
    'SecondOrderExpansion': '2nd-Order',
}

def extractJobFeatures(sub_folders):
    group_folders = {}
    for sub_folder in sub_folders:
        names = sub_folder.split('-')
        bs = [name for name in names if name.startswith('bs')][0][2:]
        qu = [name for name in names if name.startswith('qu')][0][2:]
        lpca = [name for name in names if name.startswith('lpca')][0][4:]
        train = [name for name in names if name.startswith('train')][0][5:]
        encode = [name for name in names if name.startswith('enc')][0][3:]
        depths = [name for name in names if name.startswith('Dep')][0][3:]
        QNNname = [name for name in names if name.startswith('QNN')][0][3:]
        layers = [name for name in names if name.startswith('L')][0][1:]
        loss = [name for name in names if name.startswith('l')][1][1:]
        group_name = f'{bs}-{qu}-{lpca}-{train}-{encode}-{depths}-{QNNname}-{layers}-{loss}'
        if group_name in group_folders:
            group_folders[group_name].append(sub_folder)
        else:
            group_folders[group_name] = [sub_folder]

    pprint(group_folders)
    print('\n\n')
    return group_folders

def extractAUCvalues(group_folders, catagories, skip=True):
    rows_list = []
    
    keywords = '{} {} AUC values'
    for jobfeature in sorted(group_folders):
        folders = group_folders[jobfeature]
        jobfeature = jobfeature.split('-')
        for folder in folders:
            logfile = glob.glob(f'{folder}/*log')[0]
            if skip and os.stat(logfile).st_size == 0:
                print(f'\n\033[91mNo log in: {logfile}\033[0m')
                continue
            with open(logfile) as f:
                for (i, j) in catagories:
                    dic = {
                        'batchsize': int(jobfeature[0]),
                        'Nqubits': int(jobfeature[1]),
                        'Ntrain': int(jobfeature[2]),
                        'encoder': str(Encoder_map[jobfeature[3]]),
                        'Ndepths': int(jobfeature[4]),
                        'QNNname': str(jobfeature[5]),
                        'Nlayers': int(jobfeature[6]),
                        'Events': None,
                        'AUC':  0,
                        'TimeEncode': 0,
                        'TimeTrain': 0,
                    }
                    # fill AUC value to dic
                    dic['Events'] = f'{j} {i}'
                    try:
                        f.seek(0)
                        dic['AUC'] = float(([line for line in f if keywords.format(i, j) in line][0]).split(' ')[12])

                        f.seek(0)
                        startencoding = datetime.datetime.strptime(([line for line in f if 'Start data encoding' in line][0]).split(',')[0], '%Y-%m-%d %H:%M:%S')
                        f.seek(0)
                        stopencoding = datetime.datetime.strptime(([line for line in f if 'Finish data encoding' in line][0]).split(',')[0], '%Y-%m-%d %H:%M:%S')
                        dic['TimeEncode'] = int((stopencoding-startencoding).total_seconds())

                        f.seek(0)
                        starttraining = datetime.datetime.strptime(([line for line in f if 'Train quantum model' in line][0]).split(',')[0], '%Y-%m-%d %H:%M:%S')
                        f.seek(0)
                        stoptraining = datetime.datetime.strptime(([line for line in f if 'Evaluation of quantum model' in line][0]).split(',')[0], '%Y-%m-%d %H:%M:%S')
                        dic['TimeTrain'] = int((stoptraining-starttraining).total_seconds())
                    except:
                        print(f'\n\033[91mNo AUC for {keywords.format(i, j)} in: {logfile}\033[0m')
                    rows_list.append(dic)
    df = pd.DataFrame(rows_list)
    return df

def main(args):
    input_folder = args.input_folder
    catagory = args.catagory
    catagories = [(i, j) for j in ['QNN', 'DNN'] for i in catagory]
    print(catagories)
    color = 'RdBu'
    sns.set_style('whitegrid')

    for j, top_folder in enumerate(input_folder):
        # extract physics process from top level folder
        process = [name for name in process_map.keys() if name in top_folder][0]
        print(f'\n\033[92mPhysics process: {process} ({process_map[process]})\033[0m')

        # extract all jobs
        sub_folders = glob.glob(f'{top_folder}/2020*')

        # extract job features
        group_folders = extractJobFeatures(sub_folders)

        # extract AUC values to a dataframe
        df = extractAUCvalues(group_folders, catagories)
        xTitle, yTitle, lTitle = 'Architecture', 'AUC', 'Events'
        df[xTitle] = df['encoder'] + ' ' + df['Ndepths'].apply(str) + '-Dep' + '\n' + df['QNNname'] + ' ' + df['Nlayers'].apply(str) + '-Layer'

        # get a list of names
        values=df['Ntrain'].unique().tolist()

        # create different plots for different Ntrain
        for value in values:
            sub_df = df.loc[df['Ntrain']==value]

            # calculate averaged timing
            timedf = sub_df.groupby(xTitle, as_index=False)[['TimeEncode', 'TimeTrain']].mean().set_index(xTitle)
            for i in sub_df.index:
                sub_df.at[i, xTitle] = sub_df.at[i, xTitle] + '\n(' + str(datetime.timedelta(seconds=ceil(timedf.loc[sub_df.at[i, xTitle], 'TimeEncode']))) + ')\n(' + str(datetime.timedelta(seconds=ceil(timedf.loc[sub_df.at[i, xTitle], 'TimeTrain']))) + ')'
            print(f'Ntrain {value}:')
            print(sub_df)

            fig, ax = plt.subplots(tight_layout=True)
            plt.title(f'{process_map[process]} ({value} sig {value} bkg training Events, {sub_df["Nqubits"].iloc[0]} Qubits)')
            chart = sns.boxplot(x=xTitle, y=yTitle, hue=lTitle, data=sub_df, palette=color, fliersize=0)
            chart.set_xticklabels(chart.get_xticklabels(), rotation=60)
            # jitter plot overlay
            sns.stripplot(x=xTitle, y=yTitle, hue=lTitle, data=sub_df, jitter=True, palette=color, alpha=0.8, dodge=True, linewidth=1, edgecolor='gray')
            handles, labels = ax.get_legend_handles_labels()
            plt.legend(handles[0:int(len(handles)/2)], labels[0:int(len(handles)/2)], framealpha=0.5)
            plt.ylim(0.4, 0.9)

            filename = f'{top_folder}/{yTitle}_{value}Events_{"".join(catagory)}_{process}.pdf'
            plt.savefig(filename)
            print(f'\n\033[92mSave file: {filename}\033[0m')


if __name__ == '__main__':

    """Get arguments from command line."""
    parser = ArgumentParser(description="Plot AUC information.")

    parser.add_argument('-i', '--input-folder', action='store', nargs='+', default=[], help='Folder of the log files (default: %(default)s)')
    parser.add_argument('-c', '--catagory', action='store', nargs='+', default=['Test'], help='Events to plot (default: %(default)s)')
    
    main(parser.parse_args())



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='quhep',
    version='0.0.1',
    entry_points = {
        'console_scripts': ['train_tfq=quhep.train.train_tfq:main',
            'averageROC=quhep.util.averageROC:main',
            'extracAUC=quhep.util.extracAUC:main',
            'inspecPQC=quhep.util.inspecPQC:main',
        ],
    },
    description='Quantum Machine Learning Framework for High Energy Physics',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='Rui Zhang',
    author_email='rui.zhang@cern.ch',
    keywords=['Quantum Machine Learning', 'Tensorflow Quantum', 'Qiskit pyTorch'],
    url='https://github.com/chnzhangrui/quhep',
    download_url='https://pypi.org/project/quhep/'
)

install_requires = [
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires, include_package_data=True)



#!/usr/bin/env python
# Rui Zhang 7.2020
# rui.zhang@cern.ch

import logging

import tensorflow as tf
import tensorflow_quantum as tfq
import numpy as np
import sympy
import cirq

'''
Develop your own Quantum Neural Network and make a link to the class at the bottom:
QuantumNeuralNetwork = YourClassName
'''
class QEfficientSU2NNHYB(object):
    def describe(self): return self.__class__.__name__
    def __init__(self, **kargs):
        
        self.m_nqubits = kargs['nqubits']
        self.m_loss = kargs['loss']
        self.m_use_quple = kargs['use_quple']
        self.m_nlayers = kargs['nlayers']

        logging.getLogger(__name__)

    def create_quantum_model(self, size):
        """Create a QNN model circuit and readout operation to go along with it."""
        data_qubits = cirq.GridQubit.rect(size[0], size[1]) 
        circuit = cirq.Circuit()

        entangler_map= [[i, i + 1] for i in range(size[1] - 1)]

        param_idx = 0
        num_parameters = (size[1]+1)*size[1]*2

        symb = sympy.symbols('θ_0:'+str(num_parameters))
        symbols = np.array(symb)
        symbols = symbols.reshape(size[1]+1, size[1], 2)

        for i, qubit in enumerate(data_qubits):
            circuit.append(cirq.rz(symbols[0][i][1])(qubit))
            circuit.append(cirq.ry(symbols[0][i][0])(qubit))
            param_idx += 2

        for d in range(self.m_nlayers):
            for src, targ in entangler_map:
                circuit.append(cirq.CNOT(data_qubits[src],data_qubits[targ]))
            for i, qubit in enumerate(data_qubits):
                circuit.append(cirq.rz(symbols[d+1][i][1])(qubit))
                circuit.append(cirq.ry(symbols[d+1][i][0])(qubit))
                param_idx += 2
        return circuit, [cirq.Z(i) for i in circuit.all_qubits()]

    def get_quantum_model(self, input_size):
        # Build the model circuit
        model_circuit, model_readout = self.create_quantum_model(size=(1, input_size))
        self.m_circuit = cirq.Circuit()
        self.m_circuit.append([model_circuit, model_readout])
        # SVGCircuit(model_circuit)
        logging.info(f'model circuit:\n{model_circuit}')

        # Build the Keras model
        # initializer = tf.keras.initializers.Constant(1. / input_size)
        model = tf.keras.Sequential(
            [
                # The input is the data-circuit, encoded as a tf.string
                tf.keras.layers.Input(shape=(), dtype=tf.string),
                # The PQC layer returns the expected value of the readout gate, range [-1,1].
                tfq.layers.PQC(model_circuit, model_readout),
                # tf.keras.layers.Dense(1, activation='sigmoid', kernel_initializer=initializer, name='output'),
                tf.keras.layers.Dense(1, activation='sigmoid', name='output'),
            ]
        )
        # model.layers[-1].trainable = False

        # starter_learning_rate = 0.02
        # end_learning_rate = 0.005
        # decay_steps = 1000
        # learning_rate_fn = tf.keras.optimizers.schedules.PolynomialDecay(
        #     starter_learning_rate,
        #     decay_steps,
        #     end_learning_rate,
        #     power=0.5)


        # step = tf.Variable(0, trainable=False)
        # boundaries = [100, 250]
        # values = [0.02, 0.01, 0.005]
        # learning_rate_fn = tf.keras.optimizers.schedules.PiecewiseConstantDecay(
        #     boundaries, values)

        if self.m_loss == 'hinge':
            def hinge_accuracy(y_true, y_pred):
                y_true = tf.squeeze(y_true) > 0.0
                y_pred = tf.squeeze(y_pred) > 0.0
                result = tf.cast(y_true == y_pred, tf.float32)
                return tf.reduce_mean(result)
                
            model.compile(loss=tf.keras.losses.Hinge(),
                optimizer=tf.keras.optimizers.Adam(),
                metrics=[hinge_accuracy],
                )
        elif self.m_loss == 'mse':
            model.compile(
                loss='mse',
                optimizer=tf.keras.optimizers.Adam(learning_rate=0.01, ),
                metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_Adam':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.Adam(learning_rate=learning_rate_fn, ),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_Adadelta':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.Adadelta(),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_Adamax':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.Adamax(),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_Nadam':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.Nadam(),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_SGD':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.SGD(learning_rate=0.01, momentum=0.3, nesterov=False),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_SGD_PolynomialDecay':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.SGD(learning_rate=learning_rate_fn),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])
        # elif self.m_loss == 'mse_RMSprop':
        #     model.compile(
        #         loss='mse',
        #         optimizer=tf.keras.optimizers.RMSprop(),
        #         metrics=['accuracy',tf.keras.metrics.AUC()])

                
        logging.info(f'model loss is {self.m_loss}')
          
        model.summary(print_fn=logging.getLogger(__name__).info)

        return model

QuantumNeuralNetwork = QEfficientSU2NNHYB



#!/usr/bin/env python
# Rui Zhang 6.2020
# rui.zhang@cern.ch

import logging
from argparse import ArgumentParser
import importlib

import tensorflow as tf
import tensorflow_quantum as tfq
import sympy
import cirq
from cirq.contrib.svg import SVGCircuit

from quple.trial_wavefunction.real_amplitudes import RealAmplitudes
from train import _train

class Train(_train):
    def describe(self): return self.__class__.__name__
    def __init__(self, **kargs):
        self.m_use_quple = kargs['use_quple']
        self.m_nlayers = kargs['nlayers'] if self.m_use_quple else None
        self.m_ndepths = kargs['ndepths'] if self.m_use_quple else None
        self.m_encoder_type = kargs['encoder_type'] if self.m_use_quple else None
        self.m_Qnetwork = kargs['Qnetwork']
        self.m_loss = kargs['loss']
        quple_par = f'{self.m_use_quple}-enc{self.m_encoder_type}-Dep{self.m_ndepths}-QNN{self.m_Qnetwork}-L{self.m_nlayers}-l{self.m_loss}' if self.m_use_quple else 'False'
        super().__init__(additiona_log=quple_par, **kargs)

    def convert_to_circuit(self, event):
        '''
        Encode variables into quantum datapoint
        Return:
            circuit of the event with qubit number of variable number
        '''
        qubits = cirq.GridQubit.rect(1, len(event))
        circuit = cirq.Circuit()
        for i, variable in enumerate(event):
            circuit.append(cirq.H(qubits[i]))
            circuit.append(cirq.rz(variable)(qubits[i]))
        return circuit

    def convert_to_circuit_with_entangle(self, event, entangle, copies=1):
        '''
        Encode variables into quantum datapoint
        Return:
            circuit of the event with qubit number of variable number
        '''
        qubits = cirq.GridQubit.rect(1, len(event))
        circuit = cirq.Circuit()
        for i, variable in enumerate(event):
            circuit.append(cirq.H(qubits[i]))
            circuit.append(cirq.rz(variable)(qubits[i]))
            if i < len(entangle):
                if self.m_entangle_gate == 'rx':
                    circuit.append(cirq.rx(entangle[i])(qubits[i]))
                elif self.m_entangle_gate == 'ry':
                    circuit.append(cirq.ry(entangle[i])(qubits[i]))
            for j in range(copies-1):
                circuit.append(cirq.rz(variable)(qubits[i]))
            circuit.append(cirq.H(qubits[i]))
        return circuit

    def convert_X_to_quantum(self, x_train, x_test, x_val):
        x_train_circ, x_test_circ, x_val_circ = None, None, None
        logging.info(f'Start data encoding')
        if (self.m_use_quple and (not self.m_entangle)):
            # define encoding circuit
            if self.m_encoder_type == 'FirstOrderExpansion':
                encoder = importlib.import_module(f'quple.data_encoding.first_order_expansion')
                cq = encoder.FirstOrderExpansion(feature_dimension=self.m_nqubits, parameter_symbol='data', copies=self.m_ndepths)
            elif self.m_encoder_type == 'SecondOrderExpansion':
                encoder = importlib.import_module('quple.data_encoding.second_order_expansion')
                cq = encoder.SecondOrderExpansion(feature_dimension=self.m_nqubits, parameter_symbol='data', copies=self.m_ndepths)
            else:
                logging.fatal(f'{self.m_encoder_type} is not supported')

            # get resolved circuits for input dataset
            x_train_circ = cq.resolve_parameters(x_train)
            x_test_circ = cq.resolve_parameters(x_test)
            x_val_circ = cq.resolve_parameters(x_val)
            
        else:
            x_train_circ = [self.convert_to_circuit_with_entangle(x[:self.m_nqubits], x[self.m_nqubits:], copies=self.m_ndepths) for x in x_train]
            x_test_circ = [self.convert_to_circuit_with_entangle(x[:self.m_nqubits], x[self.m_nqubits:], copies=self.m_ndepths) for x in x_test]
            x_val_circ = [self.convert_to_circuit_with_entangle(x[:self.m_nqubits], x[self.m_nqubits:], copies=self.m_ndepths) for x in x_val]

        logging.info(f'Finish data encoding. One example training event:\n{x_train_circ[0]}')
        x_train_tfcirc = tfq.convert_to_tensor(x_train_circ)
        x_test_tfcirc = tfq.convert_to_tensor(x_test_circ)
        x_val_tfcirc = tfq.convert_to_tensor(x_val_circ)
        logging.debug(tfq.from_tensor(x_train_tfcirc)[0])

        return x_train_tfcirc, x_test_tfcirc, x_val_tfcirc

    def train_quantum_model(self):
        x_train_tfcirc, x_test_tfcirc, x_val_tfcirc = self.convert_X_to_quantum(self.m_data.x_train, self.m_data.x_test, self.m_data.x_val)

        num_var = len(self.m_data.x_train[0])
        logging.info(f'Number of variables: {num_var}')
        # return x_train_tfcirc, self.m_data.y_train

        try:
            QNN = importlib.import_module(self.m_Qnetwork)
        except Exception as e:
            logging.fatal(e)

        qnn = QNN.QuantumNeuralNetwork(nqubits = self.m_nqubits, use_quple = self.m_use_quple, nlayers = self.m_nlayers, ndepths = self.m_ndepths, loss=self.m_loss)
        quantum_model = qnn.get_quantum_model(self.m_nqubits)

        logging.info(f'Train quantum model: epochs: {self.m_epochs}; batch size: {self.m_batch_size}; number of training events (sig + bkg): {x_train_tfcirc.shape[0]}')

        callback = tf.keras.callbacks.EarlyStopping(monitor='loss', patience=10, restore_best_weights=True)
        # print_weights = tf.keras.callbacks.LambdaCallback(on_epoch_end=lambda batch, logs: print(quantum_model.layers[-1].get_weights()))
        qnn_history = quantum_model.fit(
            x_train_tfcirc, self.m_data.y_train,
            batch_size=self.m_batch_size,
            epochs=self.m_epochs,
            verbose=1,
            validation_data=(x_val_tfcirc, self.m_data.y_val),
            callbacks=[callback])
        logging.info(f'Total number of QNN epochs: {len(qnn_history.history["loss"])}')

        # quantum_model.save('m.h5')
        logging.info(f'Cannot save quantum model due to a bug in TF: NotImplementedError: Layers with arguments in `__init__` must override `get_config`.')

        with open(f'{self.m_log_dir["result"]}/qnn_history.txt', 'w') as f:
            f.write( str(qnn_history.history) )

        if qnn_history is None:
            with open(f'{self.m_log_dir["result"]}/qnn_history.txt', 'r') as f:
                qnn_history = eval(r.read())

        self.plotLoss(qnn_history, 'QNN')
        qnn_results = quantum_model.evaluate(x_test_tfcirc, self.m_data.y_test)
        logging.info(f'Evaluation of quantum model on test sample: {qnn_results}')

        self.m_train_predict_QNN = quantum_model.predict(x_train_tfcirc)
        self.m_test_predict_QNN = quantum_model.predict(x_test_tfcirc)
        self.m_val_predict_QNN = quantum_model.predict(x_val_tfcirc)


    def train_classical_model(self):
        num_var = len(self.m_data.x_train[0])
        try:
            DNN = importlib.import_module('ClassicalNN')
        except Exception as e:
            logging.fatal(e)

        classical_model = DNN.get_quantum_model(num_var)

        logging.info(f'Train classical model: epochs: {self.m_epochs}; batch size: {self.m_batch_size}; number of training events (sig + bkg): {self.m_data.x_train.shape[0]}')
        callback = tf.keras.callbacks.EarlyStopping(monitor='loss', patience=10, restore_best_weights=True)
        cnn_history = classical_model.fit(self.m_data.x_train,
                self.m_data.y_train,
                batch_size=self.m_batch_size,
                epochs=self.m_epochs,
                verbose=2,
                validation_data=(self.m_data.x_val, self.m_data.y_val),
                callbacks=[callback])
        classical_model.summary(print_fn=logging.getLogger(__name__).info)
        logging.info(f'Total number of DNN epochs: {len(cnn_history.history["loss"])}')

        classical_model.save(f'{self.m_log_dir["model"]}/classical_model.h5')
        logging.info(f'Save classical model to: {self.m_log_dir["model"]}/classical_model.h5')
        
        with open(f'{self.m_log_dir["result"]}/cnn_history.txt', 'w') as f:
            f.write( str(cnn_history.history) )

        if cnn_history is None:
            with open(f'{self.m_log_dir["result"]}/cnn_history.txt', 'r') as f:
                cnn_history = eval(r.read())

        self.plotLoss(cnn_history, 'DNN')
        cnn_results = classical_model.evaluate(self.m_data.x_test, self.m_data.y_test)
        logging.info(f'Evaluation of classical model on test sample: {cnn_results}')

        self.m_train_predict_DNN = classical_model.predict(self.m_data.x_train)
        self.m_test_predict_DNN = classical_model.predict(self.m_data.x_test)
        self.m_val_predict_DNN = classical_model.predict(self.m_data.x_val)



def main(args):
    print(vars(args))
    if args.fix_seed:
        import numpy as np
        np.random.seed(args.fix_seed + 666)

    entriesToRemove = ['run_classical']
    options = vars(parser.parse_args())
    for k in entriesToRemove:
        options.pop(k, None)

    trainer = Train(**options)
    trainer.get_classical_data()

    trainer.train_quantum_model()
    if args.run_classical:
        trainer.train_classical_model()
        trainer.plotROC()
    trainer.plotROC(True)
    logging.info(f'Job finished')

if __name__ == '__main__':

    """Get arguments from command line."""
    parser = ArgumentParser(description="Tensorflow Quantum model training.")

    parser.add_argument('--batch-size', action='store', type=int, default=10, help='Training batch size type (default: %(default)s)')
    parser.add_argument('--run-classical', action='store_true', default=False, help='Train classical NN as well (default: %(default)s)')
    parser.add_argument('--entangle', action='store', type=int, default=0, help='Additinal dimension used for entanglement encoder (default: %(default)s)')
    parser.add_argument('--entangle-gate', action='store', type=str, default='rx', choices=['rx', 'ry'], help='Additinal dimension used for entanglement encoder (default: %(default)s)')
    parser.add_argument('--epochs', action='store', type=int, default=200, help='Training epochs (default: %(default)s)')
    parser.add_argument('--fix-seed', action='store', type=int, default=None, help='Fix random seed (default: %(default)s)')
    parser.add_argument('-i', '--input-file', action='store', default='../data/hmumu_twojet_100_1.npz', help='Input data (.npz) (default: %(default)s)')
    parser.add_argument('--large-pca', action='store_true', default=True, help='Use more events for PCA (default: %(default)s)')
    parser.add_argument('--log-file', action='store', default='QNN', help='Directory name to store logs (default: %(default)s)')
    parser.add_argument('--loss', action='store', default='hinge', help='Loss for QEfficientSU2NNHYB (default: %(default)s)')
    parser.add_argument('--nqubits', action='store', type=int, default=None, required=True, help='Number of qubits / varibles (default: %(default)s)')
    parser.add_argument('--output-folder', action='store', default='../output', help='Output folder for logs (default: %(default)s)')
    parser.add_argument('--padding', action='store', type=float, choices=[-999, 0], default=-999, help='Padding of non-existing (default: %(default)s)')
    parser.add_argument('--plot-input', action='store_true', default=False, help='Plotting before and after Dimension Reduction (default: %(default)s)')
    parser.add_argument('--plot-input-only', action='store_true', default=False, help='Inspect inputs (default: %(default)s)')
    parser.add_argument('--training-size', action='store', type=int, default=50, help='Number of events for training (default: %(default)s)')
    parser.add_argument('--test-size', action='store', type=int, default=None, help='Number of events for test (default: %(default)s)')
    parser.add_argument('--val-size', action='store', type=int, default=50, help='Number of events for val (default: %(default)s)')
    parser.add_argument('--Qnetwork', type=str, choices=['QDensNN', 'QConvNN', 'QConvNNHYB', 'QZinnerCNOTNN', 'QEfficientSU2NNHYB', 'QEfficientSU2NNHYB2', 'QEfficientSU2NNQNN', 'QMNISTNN'], help='Quantum network type (default: %(default)s)')
    
    subparsers = parser.add_subparsers(dest='use_quple', help='Use quple encoder')
    parser_quple = subparsers.add_parser('use_quple')
    parser_quple.add_argument('--encoder-type', type=str, required=True, choices=['FirstOrderExpansion', 'SecondOrderExpansion'], help='Data encoder depths (default: %(default)s)')
    parser_quple.add_argument('--ndepths', type=int, default=2, help='Data encoder depths (default: %(default)s)')
    parser_quple.add_argument('--nlayers', type=int, default=1, help='Number of layers for quantum model (default: %(default)s)')

    main(parser.parse_args())
    

